--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library random_lib;
use random_lib.rnd.all;

use work.axi432_pkg.all;
use work.axi432_master_pkg.all;

entity axi432_master is
  generic(cmdfile: string);         --* Name of command file
  port(clk      : in  std_ulogic;   --* clock
       srstn    : out std_ulogic;    --* Synchronous, active low, reset
       eos      : out boolean;       --* End of simulation
       hst2saxi432 : out axi432_m2s_type;  --* Host initiator to target interface
       saxi4322hst : in  axi432_s2m_type   --* Target to host initiator interface
  );

end entity axi432_master;

architecture arc of axi432_master is

  signal hst2saxi432_local:   axi432_m2s_type; --* Local host initiator to target interface

  signal srstn_local:      std_ulogic; --* Local reset

begin

  srstn     <= srstn_local;
  hst2saxi432  <= hst2saxi432_local;

  process

    type state_type is (parsing, resetting, waiting_write, waiting_read, ending);

    variable state:        state_type;
    variable are:          axi432_read_master_emulator;
    variable awe:          axi432_write_master_emulator;
    variable cnt:          integer;
    variable good:         boolean;
    variable l:            line;
    variable ln:           natural;
    variable cn:           natural;
    variable eof:          boolean;
    variable cmd:          cmd_type;
    variable val:          std_ulogic;
    variable err:          std_ulogic;

    file cf:               text;

  begin

    -- Initializations
    hst2saxi432_local <= axi432_m2s_none;
    srstn_local    <= '0';
    eos            <= false;
    free(are);
    free(awe);
    state    := parsing;
    cnt      := 1000;
    good     := true;
    file_open(cf, cmdfile, read_mode);
    l   := null;
    ln  := 1;
    cn  := 1;
    eof := false;
    cmd := RS;

    for i in 1 to 10 loop -- 10 clock cycles reset
      wait until rising_edge(clk);
    end loop;
    srstn_local <= '1';
    wait until rising_edge(clk);

    -- Main loop
    loop

      -- axi432 read address channel requests
      if hst2saxi432_local.ra.arvalid = '1' and saxi4322hst.ra.arready = '1' then -- If request acknowledged
        hst2saxi432_local.ra.arvalid <= '0'; -- Stop requesting
      end if;
      if (hst2saxi432_local.ra.arvalid = '0' or saxi4322hst.ra.arready = '1') and are.raccnt /= 0 and int_rnd(1, 10) /= 1 then
        play(are, hst2saxi432_local.ra); -- Submit next request and pop it out of requests FIFO
      end if;

      -- axi432 write address channel requests
      if hst2saxi432_local.wa.awvalid = '1' and saxi4322hst.wa.awready = '1' then -- If request acknowledged
        hst2saxi432_local.wa.awvalid <= '0'; -- Stop requesting
      end if;
      if (hst2saxi432_local.wa.awvalid = '0' or saxi4322hst.wa.awready = '1') and awe.waccnt /= 0 and int_rnd(1, 10) /= 1 then
        play(awe, hst2saxi432_local.wa); -- Submit next request and pop it out of requests FIFO
      end if;

      -- axi432 write data channel requests
      if hst2saxi432_local.wd.wvalid = '1' and saxi4322hst.wd.wready = '1' then -- If request acknowledged
        hst2saxi432_local.wd.wvalid <= '0'; -- Stop requesting
      end if;
      if (hst2saxi432_local.wd.wvalid = '0' or saxi4322hst.wd.wready = '1') and awe.wdccnt /= 0 then
        play(awe, hst2saxi432_local.wd); -- Submit next request and pop it out of requests FIFO
      end if;

      -- axi432 read data acknoledge

        hst2saxi432_local.rd.rready <= '1'; -- Randomly acknowledge responses


      if saxi4322hst.rd.rvalid = '1' and hst2saxi432_local.rd.rready = '1' then -- If response acknowledged
        check(are, saxi4322hst.rd); -- Check reponse pop it out of responses FIFO
      end if;


      -- axi432 write response acknoledge

        hst2saxi432_local.wr.bready <= '1'; -- Randomly acknowledge responses


      if saxi4322hst.wr.bvalid = '1' and hst2saxi432_local.wr.bready = '1' then -- If response acknowledged
        check(awe, saxi4322hst.wr); -- Check reponse pop it out of responses FIFO
      end if;

      -- Reset
      if srstn_local = '0' then
        free(are);
        free(awe);
      end if;

      -- State machine
      case state is
        when resetting =>
          cnt := cnt - 1;
          if cnt = 0 then
            srstn_local <= '1';
            state := parsing;
          end if;

        when ending =>
          is_empty(are, good);
          if good then
            is_empty(awe, good);
            exit when good;
          end if;

        when waiting_write =>
          cnt := cnt - 1;
          is_empty(awe, good);
          if good then
            state := parsing;
          elsif cnt = 0 then
            assert false report "Timeout exceeded while waiting for end of target axi432 transactions" severity failure;
          end if;
        
        when waiting_read =>
          cnt := cnt - 1;
          is_empty(are, good);
          if good then
            state := parsing;
          elsif cnt = 0 then
            assert false report "Timeout exceeded while waiting for end of target axi432 transactions" severity failure;
          end if;

        when parsing => -- Parsing of command file
          while state = parsing loop
            get_cmd_token(cf, l, ln, cn, cmd, eof); -- Read next command
            if eof then -- End of command file
              state := ending;
              exit;
            end if;
            case cmd is
              when RS =>
                get_dec_token(cf, l, ln, cn, cnt);
                state := resetting;
                srstn_local <= '0';
                exit;
              when ER =>
              -- Wait end of read
                get_dec_token(cf, l, ln, cn, cnt);
                state := waiting_read;
                exit;
              when EW =>
              -- Wait end of write
                get_dec_token(cf, l, ln, cn, cnt);
                state := waiting_write;
                exit;
              when WR =>
                read_vw(cf, l, ln, cn, awe);
              when RD =>
                read_vr(cf, l, ln, cn, are);
              when QT =>
                state := ending;
                exit;
            end case;
          end loop;
      end case;
      wait until rising_edge(clk);
    end loop;
    eos <= true;
    wait;

  end process;

end architecture arc;
