--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

library ieee;
use ieee.std_logic_1164.all;

library global_lib;
use global_lib.numeric_std.all;

library memories_lib;
use memories_lib.vmem_pkg;

use work.axi432_pkg.all;

-- 2**(na-2) x 32 bits register with 64-bits debug output: GPO(63 downto 32) = R(0), GPO(31 downto 0) = R(1)
entity axi432_slave is
  generic(
    s: positive range 2 to 6 := 6 -- bit-width of address bus: 1, 2, 4, 8 or 16 registers
  );
  port(
    clk:         in  std_ulogic;
    srstn:       in  std_ulogic;
    eos: 	 in  boolean;
    axi_m2s:     in  axi432_m2s_type;
    axi_s2m:     out axi432_s2m_type;
    gpo:         out std_ulogic_vector(63 downto 0)
  );
end entity axi432_slave;

architecture rtl of axi432_slave is

 -- type axi432_data_vector is array(natural range <>) of std_ulogic_vector(31 downto 0);
 -- signal r: axi432_data_vector(0 to 2**(na - 2) - 1);
  -- idle: waiting for AXI master requests: when receiving write address and data valid (higher priority than read), assert write address and data readies and
  --       go to w1, else, when receiving address read valid, assert read address ready and go to r1
  -- w1:   perform write operation, deassert write address and data readies, assert write response valid, go to w2
  -- w2:   waiting for write response ready: when receiving it, deassert write response valid, go to idle
  -- r1:   perform read operation, deassert read address ready, assert read response valid, go to r2
  -- r2:   waiting for read response ready: when receiving it, deassert read response valid, go to idle
  type state_type is (idle, w1, w2, r1, r2);
  signal state: state_type;
  constant l2bw : positive :=2;
  constant n : positive :=32;
  constant bw: positive := 2**l2bw;
  package my_vmem_pkg is new vmem_pkg
    generic map(
      l2bw => l2bw,
      l2nset => 10,
      na => 32,
      max_size => 2**20
    );
  shared variable vm: my_vmem_pkg.vmem;

begin
  vm.init;
 -- gpo <= r(0) & r(1);
   
  process(clk)
    variable add: natural range 0 to 2**(32 - 2) - 1;
  begin
    
    if rising_edge(clk) then
      if srstn = '0' then
   --     r <= (others => (others => '0'));
        axi_s2m <= axi432_s2m_none;
        state <= idle;
      else
        case state is
          when idle =>
            if axi_m2s.wa.awvalid = '1' and axi_m2s.wd.wvalid = '1' then
              axi_s2m.wa.awready <= '1';
              axi_s2m.wd.wready <= '1';
              state <= w1;
            elsif axi_m2s.ra.arvalid = '1' then
              axi_s2m.ra.arready <= '1';
              state <= r1;
            end if;
          when w1 =>
            add := to_integer(unsigned(axi_m2s.wa.awaddr(32 - 1 downto 2)));
	    vm.write(axi_m2s.wa.awaddr, axi_m2s.wd.wstrb, axi_m2s.wd.wdata);
        --    for i in 0 to 3 loop
          --    if axi_m2s.wd.wstrb(i) = '1' then
	
            --    r(add)(8 * i + 7 downto 8 * i) <= axi_m2s.wd.wdata(8 * i + 7 downto 8 * i);
             -- end if;
           -- end loop;
            axi_s2m.wa.awready <= '0';
            axi_s2m.wd.wready <= '0';
            axi_s2m.wr.bvalid <= '1';
	    axi_s2m.wr.bresp <= (others => '0');
	    axi_s2m.wr.bid <= axi_m2s.wa.awid;
            state <= w2;
          when w2 =>
            if axi_m2s.wr.bready = '1' then
              axi_s2m.wr.bvalid <= '0';
              state <= idle;
            end if;
          when r1 =>
            add := to_integer(unsigned(axi_m2s.ra.araddr(32 - 1 downto 2)));
          --  axi_s2m.rd.rdata <= r(add);
	    axi_s2m.rd.rdata <= vm.read(axi_m2s.ra.araddr);
            axi_s2m.ra.arready <= '0';
            axi_s2m.rd.rvalid <= '1';
	    axi_s2m.rd.rid <= axi_m2s.ra.arid;
	    axi_s2m.rd.rlast <= '1';
            state <= r2;
          when r2 =>
            if axi_m2s.rd.rready = '1' then
              axi_s2m.rd.rvalid <= '0';
              state <= idle;
            end if;
        end case;
      end if;
    end if;
  end process;
  vm.free;
end architecture rtl;
