--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library random_lib;
use random_lib.rnd.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.sim_utils.all;

package axi4lite32_pkg is

  subtype axi4lite32_addr_type is std_ulogic_vector(31 downto 0);
  subtype axi4lite32_wstrb_type is std_ulogic_vector(3 downto 0);
  subtype axi4lite32_prot_type is std_ulogic_vector(2 downto 0);
  subtype axi4lite32_resp_type is std_ulogic_vector(1 downto 0);
  constant axi4lite32_resp_okay: axi4lite32_resp_type := "00";
  constant axi4lite32_resp_exokay: axi4lite32_resp_type := "01";
  constant axi4lite32_resp_slverr: axi4lite32_resp_type := "10";
  constant axi4lite32_resp_decerr: axi4lite32_resp_type := "11";
  subtype axi4lite32_data_type is std_ulogic_vector(31 downto 0);

  -- Master to slave. Useful address bits are LSBs with:
  --   - 3 LSBs ignored
  --   - number of used bits depend on range of address space
  type axi4lite32_m2s_type is record
    awprot, arprot: axi4lite32_prot_type;
    awvalid, wvalid, arvalid, rready, bready: std_ulogic;
    wdata: axi4lite32_data_type;
    wstrb: axi4lite32_wstrb_type;
    awaddr, araddr: axi4lite32_addr_type;
  end record;
  constant axi4lite32_m2s_none: axi4lite32_m2s_type := (awprot => (others => '0'), arprot => (others => '0'), wdata => (others => '0'),
    wstrb => (others => '0'), awaddr => (others => '0'), araddr => (others => '0'), others => '0');

  -- Slave to master
  type axi4lite32_s2m_type is record
    awready, wready, bvalid, arready, rvalid: std_ulogic;
    bresp, rresp: axi4lite32_resp_type;
    rdata: axi4lite32_data_type;
  end record;
  constant axi4lite32_s2m_none: axi4lite32_s2m_type := (bresp => axi4lite32_resp_okay, rresp => axi4lite32_resp_okay, rdata => (others => '0'), others => '0');

  procedure mask(d: inout axi4lite32_data_type; n: axi4lite32_data_type; b: axi4lite32_wstrb_type);
  procedure mask_s(signal d: inout axi4lite32_data_type; n: axi4lite32_data_type; b: axi4lite32_wstrb_type);

  impure function axi4lite32_addr_rnd return axi4lite32_addr_type;
  impure function axi4lite32_wstrb_rnd return axi4lite32_wstrb_type;
  impure function axi4lite32_prot_rnd return axi4lite32_prot_type;
  impure function axi4lite32_resp_rnd return axi4lite32_resp_type;
  impure function axi4lite32_data_rnd return axi4lite32_data_type;
  impure function axi4lite32_m2s_rnd return axi4lite32_m2s_type;
  impure function axi4lite32_s2m_rnd return axi4lite32_s2m_type;
  procedure write(l: inout line; r: in axi4lite32_m2s_type);
  procedure write(l: inout line; r: in axi4lite32_s2m_type);

  type axi4lite32_s2m_fifo_entry;
  type axi4lite32_s2m_fifo_entry_ptr is access axi4lite32_s2m_fifo_entry;
  type axi4lite32_s2m_fifo_entry is record
    s2m: axi4lite32_s2m_type;
    nxt, prv: axi4lite32_s2m_fifo_entry_ptr;
  end record axi4lite32_s2m_fifo_entry;
  type axi4lite32_s2m_fifo is record
    head, tail: axi4lite32_s2m_fifo_entry_ptr;
    cnt: natural;
  end record axi4lite32_s2m_fifo;
  procedure init(f: inout axi4lite32_s2m_fifo);
  procedure free(f: inout axi4lite32_s2m_fifo);
  procedure push(f: inout axi4lite32_s2m_fifo; s2m: in axi4lite32_s2m_type);
  procedure pop(f: inout axi4lite32_s2m_fifo; s2m: out axi4lite32_s2m_type);

  type axi4lite32_m2s_fifo_entry;
  type axi4lite32_m2s_fifo_entry_ptr is access axi4lite32_m2s_fifo_entry;
  type axi4lite32_m2s_fifo_entry is record
    m2s: axi4lite32_m2s_type;
    nxt, prv: axi4lite32_m2s_fifo_entry_ptr;
  end record axi4lite32_m2s_fifo_entry;
  type axi4lite32_m2s_fifo is record
    head, tail: axi4lite32_m2s_fifo_entry_ptr;
    cnt: natural;
  end record axi4lite32_m2s_fifo;
  procedure init(f: inout axi4lite32_m2s_fifo);
  procedure free(f: inout axi4lite32_m2s_fifo);
  procedure push(f: inout axi4lite32_m2s_fifo; m2s: in axi4lite32_m2s_type);
  procedure pop(f: inout axi4lite32_m2s_fifo; m2s: out axi4lite32_m2s_type);

  function check(a, b: axi4lite32_m2s_type) return boolean;
  function check(a, b: axi4lite32_s2m_type) return boolean;

end package axi4lite32_pkg;

package body axi4lite32_pkg is

  procedure mask(d: inout axi4lite32_data_type; n: axi4lite32_data_type; b: axi4lite32_wstrb_type) is
  begin
    for i in 0 to 3 loop
      if b(i) = '1' then
        d(8 * i + 7 downto 8 * i) := n(8 * i + 7 downto 8 * i);
      end if;
    end loop;
  end procedure mask;

  procedure mask_s(signal d: inout axi4lite32_data_type; n: axi4lite32_data_type; b: axi4lite32_wstrb_type) is
  begin
    for i in 0 to 3 loop
      if b(i) = '1' then
        d(8 * i + 7 downto 8 * i) <= n(8 * i + 7 downto 8 * i);
      end if;
    end loop;
  end procedure mask_s;

  impure function axi4lite32_addr_rnd return axi4lite32_addr_type is
    variable res: axi4lite32_addr_type;
  begin
    res := std_ulogic_vector_rnd(axi4lite32_addr_type'length);
    return res;
  end function axi4lite32_addr_rnd;

  impure function axi4lite32_wstrb_rnd return axi4lite32_wstrb_type is
    variable res: axi4lite32_wstrb_type;
  begin
    res := std_ulogic_vector_rnd(axi4lite32_wstrb_type'length);
    return res;
  end function axi4lite32_wstrb_rnd;

  impure function axi4lite32_prot_rnd return axi4lite32_prot_type is
    variable res: axi4lite32_prot_type;
  begin
    res := std_ulogic_vector_rnd(axi4lite32_prot_type'length);
    return res;
  end function axi4lite32_prot_rnd;

  impure function axi4lite32_resp_rnd return axi4lite32_resp_type is
    variable res: axi4lite32_resp_type;
  begin
    res := std_ulogic_vector_rnd(axi4lite32_resp_type'length);
    return res;
  end function axi4lite32_resp_rnd;

  impure function axi4lite32_data_rnd return axi4lite32_data_type is
    variable res: axi4lite32_data_type;
  begin
    res := std_ulogic_vector_rnd(axi4lite32_data_type'length);
    return res;
  end function axi4lite32_data_rnd;

  impure function axi4lite32_m2s_rnd return axi4lite32_m2s_type is
    variable res: axi4lite32_m2s_type;
  begin
    res.awprot  := axi4lite32_prot_rnd;
    res.arprot  := axi4lite32_prot_rnd;
    res.awvalid := std_ulogic_rnd;
    res.wvalid  := std_ulogic_rnd;
    res.arvalid := std_ulogic_rnd;
    res.rready  := std_ulogic_rnd;
    res.bready  := std_ulogic_rnd;
    res.wdata   := axi4lite32_data_rnd;
    res.wstrb   := axi4lite32_wstrb_rnd;
    res.awaddr  := axi4lite32_addr_rnd;
    res.araddr  := axi4lite32_addr_rnd;
    return res;
  end function axi4lite32_m2s_rnd;

  impure function axi4lite32_s2m_rnd return axi4lite32_s2m_type is
    variable res: axi4lite32_s2m_type;
  begin
    res.awready := std_ulogic_rnd;
    res.wready  := std_ulogic_rnd;
    res.bvalid  := std_ulogic_rnd;
    res.arready := std_ulogic_rnd;
    res.rvalid  := std_ulogic_rnd;
    res.bresp   := axi4lite32_resp_rnd;
    res.rresp   := axi4lite32_resp_rnd;
    res.rdata   := axi4lite32_data_rnd;
    return res;
  end function axi4lite32_s2m_rnd;

  procedure write(l: inout line; r: in axi4lite32_m2s_type) is
  begin
    if r.awvalid = '1' then
      write(l, string'("AW @ 0x"));
      hwrite(l, r.awaddr);
      write(l, string'(" WPROT "));
      write(l, to_integer(unsigned(r.awprot)));
    else
      write(l, string'("                       "));
    end if;
    write(l, string'(" | "));
    if r.wvalid = '1' then
      write(l, string'("WDATA 0x"));
      hwrite(l, r.wdata);
      write(l, string'(" WSTRB 0x"));
      hwrite(l, r.wstrb);
    else
      write(l, string'("                           "));
    end if;
    write(l, string'(" | "));
    if r.arvalid = '1' then
      write(l, string'("AR @ 0x"));
      hwrite(l, r.araddr);
      write(l, string'(" RPROT "));
      write(l, to_integer(unsigned(r.arprot)));
    else
      write(l, string'("                       "));
    end if;
    write(l, string'(" | "));
    if r.bready = '1' then
      write(l, string'("BREADY"));
    else
      write(l, string'("      "));
    end if;
    write(l, string'(" | "));
    if r.rready = '1' then
      write(l, string'("RREADY"));
    else
      write(l, string'("      "));
    end if;
  end procedure write;

  procedure write(l: inout line; r: in axi4lite32_s2m_type) is
  begin
    if r.rvalid = '1' then
      write(l, string'("RDATA 0x"));
      hwrite(l, r.rdata);
      write(l, string'(" RRESP "));
      write(l, to_integer(unsigned(r.rresp)));
    else
      write(l, string'("                "));
    end if;
    write(l, string'(" | "));
    if r.bvalid = '1' then
      write(l, string'("BVALID "));
      write(l, string'(" BRESP "));
      write(l, to_integer(unsigned(r.bresp)));
    else
      write(l, string'("               "));
    end if;
    write(l, string'(" | "));
    if r.awready = '1' then
      write(l, string'("AWREADY"));
    else
      write(l, string'("       "));
    end if;
    write(l, string'(" | "));
    if r.wready = '1' then
      write(l, string'("WREADY"));
    else
      write(l, string'("      "));
    end if;
    write(l, string'(" | "));
    if r.arready = '1' then
      write(l, string'("ARREADY"));
    else
      write(l, string'("       "));
    end if;
  end procedure write;

  procedure init(f: inout axi4lite32_s2m_fifo) is
  begin
    if f.cnt /= 0 then
      free(f);
    end if;
    f.head := null;
    f.tail := null;
    f.cnt := 0;
  end procedure init;

  procedure free(f: inout axi4lite32_s2m_fifo) is
    variable r: axi4lite32_s2m_type;
  begin
    while f.cnt /= 0 loop
      pop(f, r);
    end loop;
  end procedure free;

  procedure push(f: inout axi4lite32_s2m_fifo; s2m: in axi4lite32_s2m_type) is
    variable p: axi4lite32_s2m_fifo_entry_ptr;
  begin
    p := new axi4lite32_s2m_fifo_entry;
    p.s2m := s2m;
    p.prv := null;
    if f.cnt = 0 then
      p.nxt := null;
      f.head := p;
      f.tail := p;
    else
      p.nxt := f.head;
      f.head.prv := p;
      f.head := p;
    end if;
    f.cnt := f.cnt + 1;
  end procedure push;

  procedure pop(f: inout axi4lite32_s2m_fifo; s2m: out axi4lite32_s2m_type) is
    variable p: axi4lite32_s2m_fifo_entry_ptr;
  begin
    assert f.cnt /= 0 report "Cannot pop empty S2M FIFO" severity failure;
    p := f.tail;
    s2m := p.s2m;
    f.tail := p.prv;
    if f.tail = null then
      f.head := null;
    else
      f.tail.nxt := null;
    end if;
    deallocate(p);
    f.cnt := f.cnt - 1;
  end procedure pop;

  procedure init(f: inout axi4lite32_m2s_fifo) is
  begin
    if f.cnt /= 0 then
      free(f);
    end if;
    f.head := null;
    f.tail := null;
    f.cnt := 0;
  end procedure init;

  procedure free(f: inout axi4lite32_m2s_fifo) is
    variable r: axi4lite32_m2s_type;
  begin
    while f.cnt /= 0 loop
      pop(f, r);
    end loop;
  end procedure free;

  procedure push(f: inout axi4lite32_m2s_fifo; m2s: in axi4lite32_m2s_type) is
    variable p: axi4lite32_m2s_fifo_entry_ptr;
  begin
    p := new axi4lite32_m2s_fifo_entry;
    p.m2s := m2s;
    p.prv := null;
    if f.cnt = 0 then
      p.nxt := null;
      f.head := p;
      f.tail := p;
    else
      p.nxt := f.head;
      f.head.prv := p;
      f.head := p;
    end if;
    f.cnt := f.cnt + 1;
  end procedure push;

  procedure pop(f: inout axi4lite32_m2s_fifo; m2s: out axi4lite32_m2s_type) is
    variable p: axi4lite32_m2s_fifo_entry_ptr;
  begin
    assert f.cnt /= 0 report "Cannot pop empty M2S FIFO" severity failure;
    p := f.tail;
    m2s := p.m2s;
    f.tail := p.prv;
    if f.tail = null then
      f.head := null;
    else
      f.tail.nxt := null;
    end if;
    deallocate(p);
    f.cnt := f.cnt - 1;
  end procedure pop;

  function check(a, b: axi4lite32_m2s_type) return boolean is
  begin
    return check(a.awprot, b.awprot) and
           check(a.arprot, b.arprot) and
           check(a.awvalid, b.awvalid) and
           check(a.wvalid, b.wvalid) and
           check(a.arvalid, b.arvalid) and
           check(a.rready, b.rready) and
           check(a.bready, b.bready) and
           check(a.wdata, b.wdata) and
           check(a.wstrb, b.wstrb) and
           check(a.awaddr, b.awaddr) and
           check(a.araddr, b.araddr);
  end function check;

  function check(a, b: axi4lite32_s2m_type) return boolean is
  begin
    return check(a.awready, b.awready) and
           check(a.wready, b.wready) and
           check(a.bvalid, b.bvalid) and
           check(a.arready, b.arready) and
           check(a.rvalid, b.rvalid) and
           check(a.bresp, b.bresp) and
           check(a.rresp, b.rresp) and
           check(a.rdata, b.rdata);
  end function check;

end package body axi4lite32_pkg;
