--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the AXI bridge/data logger module
-- author Letitia Li (@telecom-paristech.fr)
-- date : 2015-01-05 

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library axi_lib;
use axi_lib.axi_pkg.all;

entity axi_bridge_sim is
  generic(s0_cmdfile: string := "s0.dat";
          s1_cmdfile: string := "s1.dat";
          ngp:        positive := 32 -- Bit-width of synchronization busses between emulators
          );
end axi_bridge_sim;

architecture sim of axi_bridge_sim is 

	signal clk: std_ulogic;
	signal srstn: std_ulogic;
	signal srstn_lite: std_ulogic;
	signal eof: boolean;
	signal eof_lite: boolean;
  -- AXI lite slave port s0_axi
	signal s0_axi_m2s: axilite_gp_m2s;
	signal s0_axi_s2m: axilite_gp_s2m;
  -- AXI slave port s1_axi
	signal s1_axi_m2s: axi_gp_m2s;
	signal s1_axi_s2m: axi_gp_s2m;
  -- AXI master port m_axi
	signal m_axi_m2s: axi_gp_m2s;
	signal m_axi_s2m: axi_gp_s2m;
  -- GPIO (debug)
  signal gpi: std_ulogic_vector(7 downto 0);
  signal gpo: std_ulogic_vector(7 downto 0);
  -- Synchronization between emulators
  signal axilite2axi, axi2axilite: std_ulogic_vector(ngp - 1 downto 0);

	-- Clock period definitions
	constant clk_period : time := 2 ns;

begin

	-- Instantiate the Unit Under Test (UUT)
  uut: entity work.axi_bridge(rtl)
  port map(
    aclk       => clk,
    aresetn    => srstn_lite,
    srst       => '0',
    s1_axi_m2s => s1_axi_m2s,
    s1_axi_s2m => s1_axi_s2m,
    s0_axi_m2s => s0_axi_m2s,
    s0_axi_s2m => s0_axi_s2m,
    m_axi_m2s  => m_axi_m2s,
    m_axi_s2m  => m_axi_s2m,
    gpi        => gpi,
    gpo        => gpo
  );

  axilite_master: entity axi_lib.axilite_gp_master(arc)
  generic map(
    arrndmax => 0,
    awrndmax => 0,
    wrndmax  => 0,
    rrndmax  => 0,
    brndmax  => 0,
    cmdfile  => s0_cmdfile,
    ngp      => 32
  )
  port map(
    clk             => clk,
    axi_m2s         => s0_axi_m2s,
    axi_s2m         => s0_axi_s2m,
    srstn           => srstn_lite,
    gpi             => axi2axilite,
    gpo             => axilite2axi,
    eof             => eof_lite
  );

  axi_master: entity axi_lib.axi_gp_master(arc) 
  generic map(
    arrndmax => 0,
    awrndmax => 0,
    wrndmax  => 0,
    rrndmax  => 0,
    brndmax  => 0,
    cmdfile  => s1_cmdfile,
    ngp      => 32
  )
  port map(
    clk             => clk,
    axi_m2s         => s1_axi_m2s,
    axi_s2m         => s1_axi_s2m,
    srstn           => srstn,
    gpi             => axilite2axi,
    gpo             => axi2axilite,
    eof             => eof
  );

  axi_slave: entity axi_lib.axi_vmem(rtl)
  port map(
    clk     => clk,
    srstn   => srstn_lite,
    eos     => eof and eof_lite,
    axi_m2s => m_axi_m2s,
    axi_s2m => m_axi_s2m
  );

	-- Clock process definitions
  clk_process :process
  begin
    clk <= '0';
    wait for clk_period / 2;
    clk <= '1';
    wait for clk_period / 2;
    if eof_lite and eof then
      report "End of simulation";
      wait;
    end if; 
	end process;

  gpi <= "00000001";

end architecture sim;
